# -*- coding: utf-8 -*-
"""PyMzn can also be used to dynamically change a model during runtime. For
example, it can be useful to add constraints incrementally or change the solving
statement dynamically. To dynamically modify a model, you can use the class
``MiniZincModel``, providing a template model file as input which can then
be modified by adding variables and constraints, and by modifying the solve or
output statements. An instance of ``MiniZincModel`` can then be passed directly
to the ``minizinc`` function to be solved.
::

    model = pymzn.MiniZinModel('test.mzn')
    solutions = []
    for i in range(10):
        # add a new constraint and solve again
        model.constraint('arr_1[{0}] <= arr_2[{0}]'.format(i))
        solution = pymzn.minizinc(model)
        solutions.append(solution)
"""

import re
import os.path

from .templates import from_string
from ..dzn.marsh import stmt2dzn, val2dzn

from copy import deepcopy

stmt_p = re.compile('(?:^|;)\s*([^;]+)')
stmts_p = re.compile('(?:^|;)([^;]+)')
block_comm_p = re.compile('/\*.*\*/', re.DOTALL)
line_comm_p = re.compile('%.*\n')
var_p = re.compile('\s*([\s\w,\.\(\)\[\]\{\}\+\-\*/]+?):\s*(\w+)\s*(?:=\s*(.+))?\s*', re.DOTALL)
type_p = re.compile('\s*(?:int|float|set\s+of\s+[\s\w\.]+|array\[[\s\w\.]+\]\s*of\s*[\s\w\.]+)\s*')
array_type_p = re.compile('\s*array\[([\s\w\.]+(?:\s*,\s*[\s\w\.]+)*)\]\s+of\s+(.+)\s*')
output_stmt_p = re.compile('\s*output\s*\[(.+?)\]\s*(?:;)?\s*')
solve_stmt_p = re.compile('\s*solve\s*([^;]+)\s*(?:;)?\s*')


class Statement(object):
    """A statement of a MiniZincModel.

    Attributes
    ----------
    stmt : str
        The statement string.
    """
    def __init__(self, stmt):
        self.stmt = stmt

    def __str__(self):
        return self.stmt


class Comment(Statement):
    """A comment statement.

    Attributes
    ----------
    comment : str
        The comment string.
    """
    def __init__(self, comment):
        self.comment = comment
        stmt = '% {}\n'.format(comment)
        super().__init__(stmt)


class Constraint(Statement):
    """A constraint statement.

    Attributes
    ----------
    constr : str
        The content of the constraint, i.e. only the actual constraint without
        the starting 'constraint' and the ending semicolon.
    """
    def __init__(self, constr, comment=None):
        self.constr = constr
        stmt = 'constraint {};'.format(constr)
        super().__init__(stmt)


class Parameter(Statement):
    """A parameter statement.

    Attributes
    ----------
    par : (str, str) or (str, obj)
        Either a (name, type) pair or a (name, value) pair. In the latter case
        the type is automatically inferred from the value.
    assign : bool
        If True the parameter value will be assigned directly into the model,
        otherwise it will only be declared in the model and then it will have to
        be assigned in the data.
    """
    def __init__(self, *par, assign=True):
        name, par = par
        self.name = name
        self.type = None
        self.value = None
        self.assign = assign
        if isinstance(par, str):
            type_m = type_p.match(par)
            if type_m:
                self.type = par
        if not self.type:
            self.value = par
        if self.type:
            stmt = '{}: {};'.format(self.type, self.name)
        else:
            stmt = stmt2dzn(self.name, self.value, assign=assign)
        super().__init__(stmt)


class Variable(Statement):
    """A variable statement.

    Attributes
    ----------
    name : str
        The name of the variable.
    vartype : str
        The type of the variable.
    value : str
        The optional value of the variable statement.
    output : bool
        Whether the variable is an output variable.
    """
    def __init__(self, name, vartype, value=None, output=False):
        self.name = name
        self.value = value
        self.output = output

        array_type_m = array_type_p.match(vartype)
        if array_type_m:
            indexset = array_type_m.group(1)
            domain = array_type_m.group(2)
            if 'var' not in domain:
                vartype = 'array[{}] of var {}'.format(indexset, domain)
        elif 'var' not in vartype:
            vartype = 'var ' + vartype
        self.vartype = vartype

        stmt = '{}: {}'.format(vartype, name)
        if output:
            if array_type_m:
                stmt += ' :: output_array([{}])'.format(indexset)
            else:
                stmt += ' :: output_var'
        if value:
            stmt += ' = {}'.format(value)
        stmt += ';'

        super().__init__(stmt)


class ArrayVariable(Variable):
    """An array variable statement.

    Attributes
    ----------
    name : str
        The name of the variable.
    indexset : str
        The indexset of the array.
    domain : str
        The domain of the array.
    value : str
        The optional value of the variable statement.
    output : bool
        Whether the array variable is an output array.
    """
    def __init__(self, name, indexset, domain, value=None, output=False):
        self.indexset = indexset
        self.domain = domain
        vartype = 'array[{}] of var {}'.format(indexset, domain)
        super().__init__(name, vartype, value, output)


class OutputStatement(Statement):
    """An output statement.

    Attributes
    ----------
    output : str
        The content of the output statement, i.e. only the actual output without
        the starting 'output', the square brackets and the ending semicolon.
    """
    def __init__(self, output):
        self.output = output
        if output:
            stmt = 'output [{}];'.format(output)
        else:
            stmt = ''
        super().__init__(stmt)


class SolveStatement(Statement):
    """A solve statement.

    Attributes
    ----------
    solve : str
        The content of the solve statement, i.e. only the actual solve without
        the starting 'solve' and the ending semicolon.
    """

    def __init__(self, solve):
        self.solve = solve
        stmt = 'solve {};'.format(solve)
        super().__init__(stmt)


class MiniZincModel(object):
    """Mutable class representing a MiniZinc model.

    It can use a mzn file as template, add variables and constraints,
    modify the solve and output statements. The output statement can also be
    replaced by a dzn representation of a list of output variables.
    The final model is a string combining the existing model (if provided)
    and the updates performed on the MiniZincModel instance.

    Parameters
    ----------
    mzn : str or MiniZincModel
        A string with the content or the path to the template mzn file. If mzn
        is instead a MiniZincModel it is cloned.
    """
    def __init__(self, mzn=None):
        if mzn and isinstance(mzn, MiniZincModel):
            # if mzn is a MiniZincModel, clone it
            self.__dict__ = deepcopy(mzn.__dict__)
        else:
            self._statements = []
            self._solve_stmt = None
            self._output_stmt = None
            self._arrays = []
            self._modified = False
            self._output_vars = None

            self.mzn_file = None
            self.model = None
            if mzn and isinstance(mzn, str):
                if mzn.endswith('mzn'):
                    if os.path.isfile(mzn):
                        self.mzn_file = mzn
                    else:
                        raise ValueError('The provided file does not exsist.')
                else:
                    self.model = mzn

    def comment(self, comment):
        """Add a comment to the model.

        Parameters
        ----------
        comment : str
            The comment string.
        """
        if not isinstance(comment, Comment):
            comment = Comment(comment)
        self._statements.append(comment)
        self._modified = True

    def parameter(self, *par, assign=True):
        """Adds a parameter to the model.

        Parameters
        ----------
        par : Parameter or (str, str) or (str, obj)
            Either a Parameter, a (name, type) pair or a (name, value) pair. In
            the latter case, the type is inferred automatically from the value.
        assign : bool
            If True the parameter value will be assigned directly into the
            model, otherwise it will only be declared in the model and then it
            will have to be assigned in the data.
        """
        if isinstance(par[0], Parameter):
            par = par[0]
        else:
            par = Parameter(*par, assign=assign)
        self._statements.append(par)
        self._modified = True

    def parameters(self, pars, assign=True):
        """Add a list of parameters.

        Parameters
        ----------
        pars : list of (name, val) or Parameter
            The list of parameters to add to the model.
        assign : bool
            If True the parameters value will be assigned directly into the
            model, otherwise they will only be declared in the model and then
            they will have to be assigned in the data.
        """
        for par in pars:
            self.parameter(*par, assign=assign)
        self._modified = True

    def variable(self, name, vartype, value=None, output=False):
        """Adds a variable to the model.

        Parameters
        ----------
        name : str
            The name of the variable.
        vartype : str
            The type of the variable.
        value : str
            The optional value of the variable statement.
        output : bool
            Whether the variable is an output variable. This option is only used
            to force the compiled minizinc model to include the output_var or
            output_array annotation for this variable.
        """
        value = val2dzn(value) if value is not None else None
        array_type_m = array_type_p.match(vartype)
        if array_type_m:
            indexset = array_type_m.group(1)
            domain = array_type_m.group(2)
            dim = len(indexset.split(','))
            self._arrays.append((name, dim))
            var = ArrayVariable(name, indexset, domain, value, output)
        else:
            var = Variable(name, vartype, value, output)
        self._statements.append(var)
        self._modified = True

    def array_variable(self, name, indexset, domain, value=None, output=False):
        """Adds an array variable to the model.

        Parameters
        ----------
        name : str
            The name of the array.
        indexset : str
            The indexset of the array.
        domain : str
            The domain of the array.
        value : str
            The optional value of the array variable statement.
        output : bool
            Whether the array variable is an output array. This option is only
            used to force the compiled minizinc model to include the
            output_array annotation for this variable.
        """
        value = val2dzn(value) if value is not None else None
        var = ArrayVariable(name, indexset, domain, value, output)
        self._statements.append(var)
        self._arrays.append((var, len(indexset.split(','))))
        self._modified = True

    def constraint(self, constr):
        """Adds a constraint to the current model.

        Parameters
        ----------
        constr : str or Constraint
            As a string, the content of the constraint, i.e. only the actual
            constraint without the starting 'constraint' and the ending
            semicolon.
        """
        if not isinstance(constr, Constraint):
            constr = Constraint(constr)
        self._statements.append(constr)
        self._modified = True

    def constraints(self, constrs):
        """Add a list of constraints.

        Parameters
        ----------
        constrs : list of str or Constraint
            A list of constraints.
        """
        for constr in constrs:
            self.constraint(constr)

    def solve(self, solve_stmt):
        """Updates the solve statement of the model.

        Parameters
        ----------
        solve_stmt : str
            The content of the solve statement, i.e. only the actual solve
            without the starting 'solve' and the ending semicolon.
        """
        if not isinstance(solve_stmt, SolveStatement):
            solve_stmt = SolveStatement(solve_stmt)
        self._solve_stmt = solve_stmt
        self._modified = True

    def satisfy(self):
        """Shorthand for solve('satisfy')"""
        self._solve_stmt = SolveStatement('satisfy')
        self._modified = True

    def maximize(self, expr):
        """Shorthand for solve('maximize ' + expr)"""
        self._solve_stmt = SolveStatement('maximize ' + expr)
        self._modified = True

    def minimize(self, expr):
        """Shorthand for solve('minimize ' + expr)"""
        self._solve_stmt = SolveStatement('minimize ' + expr)
        self._modified = True

    def output(self, output_stmt):
        """Updates the output statement of the model.

        Parameters
        ----------
        solve_stmt : str
            The content of the output statement, i.e. only the actual output
            without the starting 'output', the square brackets and the ending
            semicolon.
        """
        if not isinstance(output_stmt, OutputStatement):
            output_stmt = OutputStatement(output_stmt)
        self._output_stmt = output_stmt
        self._modified = True

    def _load_model(self):
        if not self.model:
            if self.mzn_file:
                with open(self.mzn_file) as f:
                    self.model = f.read()
            else:
                self.model = ''
        return self.model

    def _parse_arrays(self):
        model = self._load_model()
        model = block_comm_p.sub('', model)
        model = line_comm_p.sub('', model)
        stmts = stmt_p.findall(model)
        arrays = []
        for stmt in stmts:
            if not stmt.strip():
                continue
            var_m = var_p.match(stmt)
            if var_m and not ('function' in stmt or 'predicate' in stmt):
                vartype = var_m.group(1)
                name = var_m.group(2)
                array_type_m = array_type_p.match(vartype)
                if array_type_m:
                    dim = len(array_type_m.group(1).split(','))
                    arrays.append((name, dim))
        return arrays

    def dzn_output(self, output_vars):
        """Sets the output statement to be a dzn representation of output_vars.

        Parameters
        ----------
        output_vars : list of str
            The list of output variables.
        """
        if not output_vars:
            return
        self._output_vars = output_vars

    def _make_dzn_output(self):
        # Parse the model to look for array declarations
        arrays = self._parse_arrays() + self._arrays

        # Build the output statement from the output variables
        out_var = '"{0} = ", show({0}), ";\\n"'
        out_array = '"{0} = array{1}d(", {2}, ", ", show({0}), ");\\n"'
        out_list = []
        for var in self._output_vars:
            if var in arrays:
                name, dim = var
                if dim == 1:
                    show_idx_sets = 'show(index_set({}))'.format(var)
                else:
                    show_idx_sets = []
                    for d in range(1, dim + 1):
                        show_idx_sets.append('show(index_set_{}of{}'
                                             '({}))'.format(d, dim, name))
                    show_idx_sets = ', ", ", '.join(show_idx_sets)
                out_list.append(out_array.format(name, dim, show_idx_sets))
            else:
                out_list.append(out_var.format(var))
        self.output(', '.join(out_list))

    def _redefine_output_vars(self, model):
        model = block_comm_p.sub('', model)
        model = line_comm_p.sub('', model)
        stmts = stmt_p.findall(model)
        modified = []
        for stmt in stmts:
            var_m = var_p.match(stmt)
            if var_m and not ('function' in stmt or 'predicate' in stmt):
                name = var_m.group(2)
                if name in self._output_vars:
                    vartype = var_m.group(1)
                    value = var_m.group(3)
                    mod = [vartype, ': ', name]
                    array_type_m = array_type_p.match(vartype)
                    if array_type_m:
                        index_set = array_type_m.group(1)
                        mod.append(' :: output_array([{}])'.format(index_set))
                    else:
                        mod.append(' :: output_var')
                    if value:
                        mod += [' = ', value]
                    mod.append(';')
                    modified.append(''.join(mod))
                    continue
            modified.append(stmt + (';' if stmt.strip() else ''))
        return '\n'.join(modified)

    @staticmethod
    def _rewrap(s):
        S = {' ', '\t', '\n', '\r', '\f', '\v'}
        stmts = []
        for stmt in stmts_p.findall(s):
            spaces = 0
            while spaces < len(stmt) and stmt[spaces] in S:
                spaces += 1
            spaces -= stmt[0] == '\n'
            lines = []
            for line in stmt.splitlines():
                start = 0
                while start < len(line) and start < spaces and line[start] in S:
                    start += 1
                lines.append(line[start:])
            stmts.append('\n'.join(lines))
        return ';\n'.join(stmts)

    def compile(self, output_file=None, args=None, rewrap=False):
        """Compiles the model and writes it to file.

        The compiled model contains the content of the template (if provided)
        plus the added variables and constraints. The solve and output
        statements will be replaced if new ones are provided.

        Parameters
        ----------
        output_file : file-like
            The file where to write the compiled model.
        args : dict
            The argumets to pass to the template engine.
        rewrap : bool
            Whether to 'prettify' the model by adjusting the indentation of the
            statements. Use only if you want to actually look at the compiled
            model.

        Returns
        -------
        str
            A string containing the generated model.
        """
        model = self._load_model()
        model = from_string(model, args)

        if rewrap:
            model = self._rewrap(model)

        if self._output_vars is not None:
            self._make_dzn_output()
            model = self._redefine_output_vars(model)

        if self._modified:
            lines = ['\n\n\n%%% GENERATED BY PYMZN %%%\n\n']

            for stmt in self._statements:
                lines.append(str(stmt))

            if self._solve_stmt:
                model = solve_stmt_p.sub('', model)
                lines.append(str(self._solve_stmt) + '\n')

            if self._output_stmt:
                model = output_stmt_p.sub('', model)
                lines.append(str(self._output_stmt) + '\n')

            model += '\n'.join(lines)

        if output_file:
            output_file.write(model)

        return model




# -*- coding: utf-8 -*-
"""
PyMzn provides functions that mimic and enhance the tools from the libminizinc
library. With these tools, it is possible to compile a MiniZinc model into a
FlatZinc one, solve a given problem and get the output solutions directly into
the python code.

The main function that PyMzn provides is the ``minizinc`` function, which
executes the entire workflow for solving a constranint program encoded in
MiniZinc.  Solving a MiniZinc problem with PyMzn is as simple as::

    import pymzn
    pymzn.minizinc('test.mzn')

The ``minizinc`` function is probably the way to go for most of the problems,
but the ``mzn2fzn`` and ``solns2out`` functions are also included in the public
API to allow for maximum flexibility. The latter two functions are wrappers of
the two homonym MiniZinc tools for, respectively, converting a MiniZinc model
into a FlatZinc one and getting custom output from the solution stream of a
solver.
"""

import os
import logging
import contextlib

from io import BufferedReader, TextIOWrapper
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile

import pymzn.config as config

from . import solvers
from .solvers import gecode
from .model import MiniZincModel
from ..process import Process
from ..dzn import dict2dzn, dzn2dict


class Solutions:
    """Represents a solution stream from the `minizinc` function.

    This class populates lazily but can be referenced and iterated as a list.

    Attributes
    ----------
    complete : bool
        Whether the stream includes the complete set of solutions. This means
        the stream contains all solutions in a satisfiability problem, or it
        contains the global optimum for maximization/minimization problems.
    """

    def __init__(self, stream):
        self._stream = stream
        self._solns = []
        self.complete = False
        self._iter = None
        self._stats = None

    @property
    def statistics(self):
        self._fetch_all()
        return self._stats

    def _fetch(self):
        try:
            solution = next(self._stream)
            self._solns.append(solution)
            return solution
        except StopIteration as stop:
            complete, stats = stop.value
            self.complete = complete
            if stats:
                self._stats = stats
            self._stream = None
        return None

    def _fetch_all(self):
        while self._stream:
            self._fetch()

    def __len__(self):
        self._fetch_all()
        return len(self._solns)

    def __next__(self):
        if self._stream:
            return self._fetch()
        else:
            if not self._iter:
                self._iter = iter(self._solns)
            try:
                return next(self._iter)
            except StopIteration:
                self._iter = iter(self._solns)
                raise

    def __iter__(self):
        if not self._stream:
            self._iter = iter(self._solns)
        return self

    def __getitem__(self, key):
        self._fetch_all()
        return self._solns[key]

    def __repr__(self):
        self._fetch_all()
        return repr(self._solns)

    def __str__(self):
        self._fetch_all()
        return str(self._solns)


def minizinc(
        mzn, *dzn_files, data=None, keep=False, include=None, solver=None,
        output_mode='dict', output_vars=None, output_dir=None, timeout=None,
        all_solutions=False, num_solutions=None, force_flatten=False, args=None,
        wait=False, statistics=False, **kwargs
    ):
    """Implements the workflow to solve a CSP problem encoded with MiniZinc.

    Parameters
    ----------
    mzn : str or MiniZincModel
        The minizinc problem to be solved.  It can be either a string or an
        instance of MiniZincModel.  If it is a string, it can be either the path
        to the mzn file or the content of the model.
    *dzn_files
        A list of paths to dzn files to attach to the mzn2fzn execution,
        provided as positional arguments; by default no data file is attached.
        Data files are meant to be used when there is data that is static across
        several minizinc executions.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the mzn2fnz function. The dictionary is then automatically converted to
        dzn format by the pymzn.dzn function. This property is meant to include
        data that dynamically changes across several minizinc executions.
    keep : bool
        Whether to keep the generated mzn, dzn, fzn and ozn files or not. If
        False, the generated files are created as temporary files which will be
        deleted right after the problem is solved. Though pymzn generated files
        are not originally intended to be kept, this property can be used for
        debugging purpose. Notice that in case of error the files are not
        deleted even if this parameter is False.  Default is False.
    include : str or list
        One or more additional paths to search for included mzn files.
    solver : Solver
        An instance of Solver to use to solve the minizinc problem. The default
        is pymzn.gecode.
    output_mode : 'dzn', 'json', 'item', 'dict'
        The desired output format. The default is 'dict' which returns a stream
        of solutions decoded as python dictionaries. The 'item' format outputs a
        stream of strings as returned by the solns2out tool, formatted according
        to the output statement of the MiniZinc model. The 'dzn' and 'json'
        formats output a stream of strings formatted in dzn of json
        respectively.
    output_vars : [str]
        A list of output variables. These variables will be the ones included in
        the output dictionary. Only available if ouptut_mode='dict'.
    output_dir : str
        Output directory for files generated by PyMzn. The default (None) is the
        temporary directory of your OS (if keep=False) or the current working
        directory (if keep=True).
    timeout : int
        Number of seconds after which the solver should stop the computation and
        return the best solution found. This is only available if the solver has
        support for a timeout.
    all_solutions : bool
        Whether all the solutions must be returned. Notice that this can only
        be used if the solver supports returning all solutions. Default is False.
    num_solutions : int
        The upper bound on the number of solutions to be returned. Can only be
        used if the solver supports returning a fixed number of solutions.
        Default is 1.
    force_flatten : bool
        Wheter the function should be forced to produce a flat model. Whenever
        possible, this function feeds the mzn file to the solver without passing
        through the flattener, force_flatten=True prevents this behavior and
        always produces a fzn file which is in turn passed to the solver.
    args : dict
        Arguments for the template engine.
    wait : bool
        Whether to wait for the solving process to finish before returning the
        solution stream. This parameter is ignored on Windows systems, on which
        the solving process is always awaited.
    statistics : bool
        Whether to save the statistics of the solver (if supported).
    **kwargs
        Additional arguments to pass to the solver, provided as additional
        keyword arguments to this function. Check the solver documentation for
        the available arguments.

    Returns
    -------
    Solutions
        Returns a list of solutions as a Solutions instance. The actual content
        of the stream depends on the output_mode chosen.
    """
    if isinstance(mzn, MiniZincModel):
        mzn_model = mzn
    else:
        mzn_model = MiniZincModel(mzn)

    if not solver:
        solver = config.get('solver', gecode)
    elif isinstance(solver, str):
        solver = getattr(solvers, solver)

    if all_solutions and not solver.support_all:
        raise ValueError('The solver cannot return all solutions')
    if num_solutions is not None and not solver.support_num:
        raise ValueError('The solver cannot return a given number of solutions')
    if output_mode != 'dict' and output_vars:
        raise ValueError('Output vars only available in `dict` output mode')
    if statistics and not solver.support_stats:
        raise ValueError('The solver does not support emitting statistics')

    if not output_dir:
        output_dir = config.get('output_dir', None)

    keep = config.get('keep', keep)

    if output_mode == 'dict':
        if output_vars:
            mzn_model.dzn_output(output_vars)
            _output_mode = 'item'
        else:
            _output_mode = 'dzn'
    else:
        _output_mode = output_mode

    output_prefix = 'pymzn'
    if keep:
        mzn_dir = os.getcwd()
        if mzn_model.mzn_file:
            mzn_dir, mzn_name = os.path.split(mzn_model.mzn_file)
            output_prefix, _ = os.path.splitext(mzn_name)
        output_dir = output_dir or mzn_dir

    output_prefix += '_'
    output_file = NamedTemporaryFile(dir=output_dir, prefix=output_prefix,
                                     suffix='.mzn', delete=False, mode='w+',
                                     buffering=1)
    mzn_model.compile(output_file, rewrap=keep, args=args)
    output_file.close()

    mzn_file = output_file.name
    data_file = None
    fzn_file = None
    ozn_file = None

    wait = wait or os.name == 'nt'

    force_flatten = (
           config.get('force_flatten', force_flatten)
        or not solver.support_mzn
        or (_output_mode in ['dzn', 'json'] and not solver.support_output_mode)
    )

    solver_args = {**config.get('solver_args', {}), **kwargs}
    try:
        if force_flatten:
            fzn_file, ozn_file = mzn2fzn(
                mzn_file, *dzn_files, data=data, keep_data=keep,
                include=include, globals_dir=solver.globals_dir,
                output_mode=_output_mode
            )
            solver_stream = _solve(
                solver, fzn_file, wait=wait, timeout=timeout, output_mode='dzn',
                all_solutions=all_solutions, num_solutions=num_solutions,
                statistics=statistics, **solver_args
            )
            out = solns2out(solver_stream, ozn_file)
        else:
            dzn_files = list(dzn_files)
            data, data_file = _prepare_data(mzn_file, data, keep)
            if data_file:
                dzn_files.append(data_file)
            out = _solve(
                solver, mzn_file, *dzn_files, wait=wait, lines=True, data=data,
                include=include, timeout=timeout, all_solutions=all_solutions,
                num_solutions=num_solutions, output_mode=_output_mode,
                statistics=statistics, **solver_args
            )
        solns = split_solns(out)
        if output_mode == 'dict':
            solns = _to_dict(solns)
        stream = solns
    except (
        MiniZincUnsatisfiableError, MiniZincUnknownError, MiniZincUnboundedError
    ) as err:
        err.mzn_file = mzn_file
        raise err

    cleanup_files = [] if keep else [data_file, mzn_file, fzn_file, ozn_file]
    stream = _cleanup(stream, cleanup_files)
    return Solutions(stream)


def _cleanup(stream, files):
    try:
        while True:
            yield next(stream)
    except StopIteration as stop:
        return stop.value
    finally:
        log = logging.getLogger(__name__)
        with contextlib.suppress(FileNotFoundError):
            for _file in files:
                if _file:
                    os.remove(_file)
                    log.debug('Deleting file: {}'.format(_file))


def _solve(solver, *args, lines=False, wait=False, **kwargs):
    if wait:
        out = solver.solve(*args, **kwargs)
        if lines:
            return out.splitlines()
        return out
    else:
        solver_process = solver.solve_start(*args, **kwargs)
        if lines:
            return solver_process.readlines()
        return solver_process


def mzn2fzn(mzn_file, *dzn_files, data=None, keep_data=False, globals_dir=None,
            include=None, output_mode='item', no_ozn=False):
    """Flatten a MiniZinc model into a FlatZinc one. It executes the mzn2fzn
    utility from libminizinc to produce a fzn and ozn files from a mzn one.

    Parameters
    ----------
    mzn_file : str
        The path to the minizinc problem file.
    *dzn_files
        A list of paths to dzn files to attach to the mzn2fzn execution,
        provided as positional arguments; by default no data file is attached.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the mzn2fnz function. The dictionary is then automatically converted to
        dzn format by the ``pymzn.dict2dzn`` function. Notice that if the data
        provided is too large, a temporary dzn file will be produced.
    keep_data : bool
        Whether to write the inline data into a dzn file and keep it.
        Default is False.
    globals_dir : str
        The path to the directory for global included files.
    include : str or list
        One or more additional paths to search for included mzn files when
        running ``mzn2fzn``.
    output_mode : 'dzn', 'json', 'item'
        The desired output format. The default is 'item' which outputs a
        stream of strings as returned by the solns2out tool, formatted according
        to the output statement of the MiniZinc model. The 'dzn' and 'json'
        formats output a stream of strings formatted in dzn of json
        respectively.
    no_ozn : bool
        If True, the ozn file is not produced, False otherwise.

    Returns
    -------
    tuple (str, str)
        The paths to the generated fzn and ozn files. If ``no_ozn=True``, the
        second argument is None.
    """
    args = [config.get('mzn2fzn', 'mzn2fzn')]
    if globals_dir:
        args += ['-G', globals_dir]
    if no_ozn:
        args.append('--no-output-ozn')
    if output_mode and output_mode in ['dzn', 'json', 'item']:
        args += ['--output-mode', output_mode]
    if include:
        if isinstance(include, str):
            include = [include]
        elif not isinstance(include, list):
            raise TypeError('The path provided is not valid.')
    else:
        include = []

    include += config.get('include', [])
    for path in include:
        args += ['-I', path]

    keep_data = config.get('keep', keep_data)

    dzn_files = list(dzn_files)
    data, data_file = _prepare_data(mzn_file, data, keep_data)
    if data:
        args += ['-D', data]
    elif data_file:
        dzn_files.append(data_file)
    args += [mzn_file] + dzn_files

    log = logging.getLogger(__name__)

    process = None
    try:
        process = Process(args).run()
    except CalledProcessError as err:
        log.exception(err.stderr)
        raise RuntimeError(err.stderr) from err

    if not keep_data:
        with contextlib.suppress(FileNotFoundError):
            if data_file:
                os.remove(data_file)
                log.debug('Deleting file: {}'.format(data_file))

    mzn_base = os.path.splitext(mzn_file)[0]
    fzn_file = '.'.join([mzn_base, 'fzn'])
    fzn_file = fzn_file if os.path.isfile(fzn_file) else None
    ozn_file = '.'.join([mzn_base, 'ozn'])
    ozn_file = ozn_file if os.path.isfile(ozn_file) else None

    if fzn_file:
        log.debug('Generated file: {}'.format(fzn_file))
    if ozn_file:
        log.debug('Generated file: {}'.format(ozn_file))

    return fzn_file, ozn_file


def _prepare_data(mzn_file, data, keep_data=False):
    if not data:
        return None, None

    if isinstance(data, dict):
        data = dict2dzn(data)
    elif isinstance(data, str):
        data = [data]
    elif not isinstance(data, list):
        raise TypeError('The additional data provided is not valid.')

    log = logging.getLogger(__name__)

    if keep_data or sum(map(len, data)) >= config.get('dzn_width', 70):
        mzn_base, __ = os.path.splitext(mzn_file)
        data_file = mzn_base + '_data.dzn'
        with open(data_file, 'w') as f:
            f.write('\n'.join(data))
        log.debug('Generated file: {}'.format(data_file))
        data = None
    else:
        data = ' '.join(data)
        data_file = None
    return data, data_file


def _solns2out_process(ozn_file):
    args = [config.get('solns2out', 'solns2out'), ozn_file]
    process = Process(args)
    return process


def solns2out(stream, ozn_file):
    """Wraps the solns2out utility, executes it on the solution stream, and
    then returns the output stream.

    Parameters
    ----------
    stream : str or BufferedReader
        The solution stream returned by the solver.
    ozn_file : str
        The ozn file path produced by the mzn2fzn function.

    Returns
    -------
    generator of str
        The output stream of solns2out encoding the solution stream according to
        the provided ozn file.
    """
    log = logging.getLogger(__name__)
    args = [config.get('solns2out', 'solns2out'), ozn_file]
    process = _solns2out_process(ozn_file)
    try:
        if isinstance(stream, (BufferedReader, TextIOWrapper)):
            process.start(stream)
            yield from process.readlines()
        elif isinstance(stream, Process):
            if stream.alive:
                process.start(stream.stdout)
                yield from process.readlines()
            else:
                process.run(stream.stdout_data)
                yield from process.stdout_data.splitlines()
        else:
            process.run(stream)
            yield from process.stdout_data.splitlines()
    except CalledProcessError as err:
        log.exception(err.stderr)
        raise RuntimeError(err.stderr) from err


SOLN_SEP = '----------'
SEARCH_COMPLETE = '=========='
UNSATISFIABLE = '=====UNSATISFIABLE====='
UNKNOWN = '=====UNKNOWN====='
UNBOUNDED = '=====UNBOUNDED====='


def split_solns(lines):
    """Split the solutions from the output stream of a solver or solns2out"""
    _buffer = []
    complete = False
    for line in lines:
        line = line.strip()
        if line == SOLN_SEP:
            yield '\n'.join(_buffer)
            _buffer = []
        elif line == SEARCH_COMPLETE:
            complete = True
            _buffer = []
        elif line == UNKNOWN:
            raise MiniZincUnknownError
        elif line == UNSATISFIABLE:
            raise MiniZincUnsatisfiableError
        elif line == UNBOUNDED:
            raise MiniZincUnboundedError
        else:
            _buffer.append(line)
    return (complete, '\n'.join(_buffer))


def _to_dict(stream):
    try:
        while True:
            yield dzn2dict(next(stream))
    except StopIteration as stop:
        return stop.value


class MiniZincError(RuntimeError):
    """Generic error for the MiniZinc functions."""

    def __init__(self, msg=None):
        super().__init__(msg)
        self._mzn_file = None

    @property
    def mzn_file(self):
        """str: the mzn file that generated the error."""
        return self._mzn_file

    @mzn_file.setter
    def mzn_file(self, _mzn_file):
        self._mzn_file = _mzn_file


class MiniZincUnsatisfiableError(MiniZincError):
    """Error raised when a minizinc problem is found to be unsatisfiable."""

    def __init__(self):
        super().__init__('The problem is unsatisfiable.')


class MiniZincUnknownError(MiniZincError):
    """Error raised when minizinc returns no solution (unknown)."""

    def __init__(self):
        super().__init__('The solution of the problem is unknown.')


class MiniZincUnboundedError(MiniZincError):
    """Error raised when a minizinc problem is found to be unbounded."""

    def __init__(self):
        super().__init__('The problem is unbounded.')




# -*- coding: utf-8 -*-

import logging
import pymzn.config as config

from textwrap import TextWrapper
from numbers import Integral, Real, Number
from collections.abc import Set, Sized, Iterable, Mapping

_wrapper = None

def _get_wrapper():
    global _wrapper
    if not _wrapper:
        _wrapper = TextWrapper(width=config.get('dzn_width', 70),
                               subsequent_indent=' '*4, break_long_words=False,
                               break_on_hyphens = False)
    return _wrapper


def _is_bool(obj):
    return isinstance(obj, bool)


def _is_int(obj):
    return isinstance(obj, Integral)


def _is_float(obj):
    return isinstance(obj, Real)


def _is_value(obj):
    return isinstance(obj, (bool, str, Number))


def _is_set(obj):
    return isinstance(obj, Set) and all(map(_is_value, obj))


def _is_elem(obj):
    return _is_value(obj) or _is_set(obj)


def _is_list(obj):
    return (isinstance(obj, Sized) and isinstance(obj, Iterable) and
            not isinstance(obj, (Set, Mapping)))


def _is_dict(obj):
    return isinstance(obj, Mapping)


def _is_array_type(obj):
    return _is_list(obj) or _is_dict(obj)


def _list_index_set(obj):
    return 1, len(obj)


def _extremes(s):
    return min(s), max(s)


def _is_int_set(obj):
    return all(map(_is_int, obj))


def _is_contiguous(obj, min_val, max_val):
    return all([v in obj for v in range(min_val, max_val + 1)])


def _index_set(obj):
    if _is_list(obj):
        if len(obj) == 0:
            return ()
        if all(map(_is_elem, obj)):
            return _list_index_set(obj),
        if all(map(_is_array_type, obj)):
            idx_sets = list(map(_index_set, obj))
            # all children index-sets must be identical
            if idx_sets[1:] == idx_sets[:-1]:
                return (_list_index_set(obj),) + idx_sets[0]
    elif _is_dict(obj):
        if len(obj) == 0:
            return ()
        keys = obj.keys()
        if _is_int_set(keys):
            min_val, max_val = _extremes(keys)
            if _is_contiguous(keys, min_val, max_val):
                idx_set = (min_val, max_val),
                if all(map(_is_elem, obj.values())):
                    return idx_set
                if all(map(_is_array_type, obj.values())):
                    idx_sets = list(map(_index_set, obj.values()))
                    # all children index-sets must be identical
                    if idx_sets[1:] == idx_sets[:-1]:
                        return idx_set + idx_sets[0]
    raise ValueError('The input object is not a proper array: '
                     '{}'.format(repr(obj)), obj)


def _flatten_array(arr, lvl):
    if _is_dict(arr):
        arr_it = arr.values()
    else:
        arr_it = arr

    if lvl == 1:
        return arr_it

    flat_arr = []
    for sub_arr in arr_it:
        flat_arr.extend(_flatten_array(sub_arr, lvl - 1))
    return flat_arr


def _dzn_val(val):
    if isinstance(val, bool):
        return 'true' if val else 'false'
    return str(val)


def _dzn_set(s):
    if s and _is_int_set(s):
        min_val, max_val = _extremes(s)
        if _is_contiguous(s, min_val, max_val):
            return '{}..{}'.format(min_val, max_val)  # contiguous set
    return '{{{}}}'.format(', '.join(map(_dzn_val, s)))


def _index_set_str(idx_set):
    return ', '.join(['{}..{}'.format(*s) for s in idx_set])


def _dzn_array_nd(arr):
    idx_set = _index_set(arr)
    dim = max([len(idx_set), 1])
    if dim > 6:  # max 6-dimensional array in dzn language
        raise ValueError('The input array has {} dimensions. Minizinc supports'
                         ' arrays of up to 6 dimensions.'.format(dim), arr)

    if _is_dict(arr):
        arr_it = arr.values()
    else:
        arr_it = arr
    flat_arr = _flatten_array(arr_it, dim)

    dzn_arr = 'array{}d({}, {})'
    if len(idx_set) > 0:
        idx_set_str = _index_set_str(idx_set)
    else:
        idx_set_str = '{}' # Empty index set
    vals = []
    for i, val in enumerate(map(_dzn_val, flat_arr)):
        if i > 0:
            vals.append(', ')
        vals.append(val)

    arr_str = '[{}]'.format(''.join(vals))
    return dzn_arr.format(dim, idx_set_str, arr_str)


def _array_elem_type(arr, idx_set):
    if len(idx_set) == 0:
        return _dzn_type(arr)

    it = iter(arr.values()) if _is_dict(arr) else iter(arr)
    return _array_elem_type(next(it), idx_set[1:])


def _dzn_type(val):
    if _is_bool(val):
        return 'bool'
    if _is_int(val):
        return 'int'
    if _is_float(val):
        return 'float'
    if _is_set(val):
        if len(val) == 0:
            raise TypeError('The given set is empty.')
        return 'set of {}'.format(_dzn_type(next(iter(val))))
    if _is_array_type(val):
        idx_set = _index_set(val)
        if len(idx_set) == 0:
            raise TypeError('The given array is empty.')
        idx_set_str = _index_set_str(idx_set)
        elem_type = _array_elem_type(val, idx_set)
        return 'array[{}] of {}'.format(idx_set_str, elem_type)
    raise TypeError('Unsupported parsing for value: {}'.format(repr(val)))


def val2dzn(val, wrap=True):
    """Serializes a value (bool, int, float, set, array) into its dzn
    representation.

    Parameters
    ----------
    val
        The value to serialize
    wrap : bool
        Whether to wrap the serialized value.

    Returns
    -------
    str
        The serialized dzn representation of the given value.
    """
    if _is_value(val):
        dzn_val = _dzn_val(val)
    elif _is_set(val):
        dzn_val = _dzn_set(val)
    elif _is_array_type(val):
        dzn_val =_dzn_array_nd(val)
    else:
        raise TypeError('Unsupported parsing for value: {}'.format(repr(val)))

    if wrap:
        wrapper = _get_wrapper()
        dzn_val = wrapper.fill(dzn_val)

    return dzn_val


def stmt2dzn(name, val, declare=True, assign=True, wrap=True):
    """Returns a dzn statement declaring and assigning the given value.

    Parameters
    ----------
    val
        The value to serialize.
    declare : bool
        Whether to include the declaration of the variable in the statement or
        just the assignment.
    assign : bool
        Wheter to include assignment of the value in the statement or just the
        declaration.
    wrap : bool
        Whether to wrap the serialized value.

    Returns
    -------
    str
        The serialized dzn representation of the value.
    """
    if not (declare or assign):
        raise ValueError('The statement must be a declaration or an assignment.')

    stmt = []
    if declare:
        val_type = _dzn_type(val)
        stmt.append('{}: '.format(val_type))
    stmt.append(name)
    if assign:
        val_str = val2dzn(val, wrap=wrap)
        stmt.append(' = {}'.format(val_str))
    stmt.append(';')
    return ''.join(stmt)


def dict2dzn(objs, declare=False, assign=True, wrap=True, fout=None):
    """Serializes the objects in input and produces a list of strings encoding
    them into dzn format. Optionally, the produced dzn is written in a file.

    Supported types of objects include: str, int, float, set, list or dict.
    List and dict are serialized into dzn (multi-dimensional) arrays. The
    key-set of a dict is used as index-set of dzn arrays. The index-set of a
    list is implicitly set to 1..len(list).

    Parameters
    ----------
    objs : dict
        A dictionary containing the objects to serialize, the keys are the names
        of the variables.
    declare : bool
        Whether to include the declaration of the variable in the statements or
        just the assignment. (Default is False for backward compatibility)
    assign : bool
        Wheter to include assignment of the value in the statements or just the
        declaration.
    wrap : bool
        Whether to wrap the serialized values.
    fout : str
        Path to the output file, if None no output file is written.

    Returns
    -------
    list
        List of strings containing the dzn-encoded objects.
    """
    log = logging.getLogger(__name__)

    vals = []
    for key, val in objs.items():
        stmt = stmt2dzn(key, val, declare=declare, assign=assign, wrap=wrap)
        vals.append(stmt)

    if fout:
        log.debug('Writing file: {}'.format(fout))
        with open(fout, 'w') as f:
            for val in vals:
                f.write('{}\n\n'.format(val))
    return vals


def rebase_array(d, recursive=False):
    """Transform an indexed dictionary (such as those returned by the dzn2dict
    function when parsing arrays) into an multi-dimensional list.

    Parameters
    ----------
    d : dict
        The indexed dictionary to convert.
    bool : recursive
        Whether to rebase the array recursively.

    Returns
    -------
    list
        A multi-dimensional list.
    """
    arr = []
    min_val, max_val = _extremes(d.keys())
    for idx in range(min_val, max_val + 1):
        v = d[idx]
        if recursive and _is_dict(v):
            v = rebase_array(v)
        arr.append(v)
    return arr




# -*- coding: utf-8 -*-
"""Handles MiniZinc template files.

PyMzn supports templating as a form of dynamic modelling. PyMzn allows to embed
code from the `Jinja2 <http://jinja.pocoo.org/>`_ templating language within a
MiniZinc model to make a PyMzn template file (usually distinguished with the
`pmzn` extension). An example::

    %% knapsack.pmzn %%
    int: n;                     % number of objects
    set of int: OBJ = 1..n;
    int: capacity;              % the capacity of the knapsack
    array[OBJ] of int: size;    % the size of each object=

    var set of OBJ: x;
    constraint sum(i in x)(size[i]) <= capacity;

    {% if objective == 'profit' %}
        array[OBJ] of int: profit;  % the profit of each object
        solve maximize sum(i in x)(profit[i]);
    {% elif objective == 'cost' %}
        array[OBJ] of int: cost;    % the cost of each object
        solve minimize sum(i in x)(cost[i]);
    {% else %}
        solve satisfy;
    {% endif%}

    %% knapsack.dzn %%
    n = 5;
    size = [14, 4, 10, 6, 9];
    capacity = 20;

Now it is possible to pass `objective` as a member of `args` in the
`pymzn.minizinc` function::

    pymzn.minizinc('knapsack.pmzn', data={'cost': [10, 3, 9, 4, 8]}, args={'objective': 'cost'})

The compiled model that will be solved looks like this::

    %% knapsack.mzn %%
    int: n;                     % number of objects
    set of int: OBJ = 1..n;
    int: capacity;              % the capacity of the knapsack
    array[OBJ] of int: size;    % the size of each object=

    var set of OBJ: x;
    constraint sum(i in x)(size[i]) <= capacity;

    array[OBJ] of int: cost;    % the cost of each object
    solve minimize sum(i in x)(cost[i]);

Notice that now the model expects a `cost` array as dzn input data, so we passed
it with the `data` argument as usual.

As mentioned, PyMzn employs Jinja2 under the hood, so anything you can do with
Jinja2 is also possible in PyMzn, including variables, control structured,
template inheritance, and filters. PyMzn implements few custom filters as well:

- `int(value, factor=100)` : discretizes the given input or array,
  pre-multiplying by the given factor.
- `dzn(value)` : transform the input into its equivalent dzn string.
"""

from copy import deepcopy
from collections.abc import Iterable
from pymzn.dzn.marsh import val2dzn
from jinja2 import (
    Environment, Template, BaseLoader, PackageLoader, FileSystemLoader,
    TemplateNotFound
)


def discretize(value, factor=100):
    """Discretize the given value, pre-multiplying by the given factor"""
    if not isinstance(value, Iterable):
        return int(value * factor)
    int_value = list(deepcopy(value))
    for i in range(len(int_value)):
        int_value[i] = int(int_value[i] * factor)
    return int_value


class MultiLoader(BaseLoader):

    def __init__(self):
        self._loaders = []

    def add_loader(self, loader):
        self._loaders.append(loader)

    def get_source(self, environment, template):
        for loader in self._loaders:
            try:
                return loader.get_source(environment, template)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(template)

    def list_templates(self):
        seen = set()
        templates = []
        for loader in self._loaders:
            for tmpl in loader.list_templates():
                if tmpl not in seen:
                    templates.append(tmpl)
                    seen.add(tmpl)
        return templates


_jload = MultiLoader()
_jenv = Environment(trim_blocks=True, lstrip_blocks=True, loader=_jload)
_jenv.filters['dzn'] = val2dzn
_jenv.filters['int'] = discretize


def from_string(source, args=None):
    """Renders a template string"""
    return _jenv.from_string(source).render(args or {})


def add_package(package_name, package_path='templates', encoding='utf-8'):
    """Adds the given package to the template search routine"""
    _jload.add_loader(PackageLoader(package_name, package_path, encoding))


def add_path(searchpath, encoding='utf-8', followlinks=False):
    """Adds the given path to the template search routine"""
    _jload.add_loader(FileSystemLoader(searchpath, encoding, followlinks))

